/*  09.05.27 - servable.nut
 *
 *  This file is part of Trans AI
 *
 *  Copyright 2009 fanio zilla <fanio.zilla@gmail.com>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 *  MA 02110-1301, USA.
 */

 /**
 * Base Serv-able class for store either town or industry handling in game
 */
class Servable extends Storable
{
    _API = null; ///< common API function to be used by this class
    constructor()
    {
        ::Storable.constructor();
        this._storage._is_town <- true; ///< Flag of town
        this._storage._area <- []; ///< Array of area tiles
        this._API = AITown;
    }

    /**
     * Get the name of servable object
     * @return string name
     */
    function GetName() { return this._API.GetName(this.GetID()); }

    /**
     * Get type of servable object
     * @return true if Town or  false if Industry
     */
    function IsTown() { return this._storage._is_town; }

    /**
     * Get the location of servable object
     * @return tile index of location
     */
    function GetLocation() { return this._API.GetLocation(this.GetID()); }

    /**
     * Get the total last month's production of the given cargo.
     * @param cargo_id  The index of the cargo.
     * @return The last month's production of the given cargo
     */
    function GetLastMonthProduction(cargo_id)
    {
        if (!AICargo.IsValidCargo(cargo_id)) return 0;
        return this._API.GetLastMonthProduction(this.GetID(), cargo_id);
    }

    /**
     * Get the total amount of cargo transported  last month.
     * @param cargo_id  The index of the cargo.
     * @return The amount of given cargo transported last month.
     */
    function GetLastMonthTransported(cargo_id)
    {
        if (!AICargo.IsValidCargo(cargo_id)) return 0;
        return this._API.GetLastMonthTransported(this.GetID(), cargo_id);
    }

    /**
     * Get the manhattan distance from the tile
     * @param tile  The tile to get the distance to.
     * @return The distance between this object and tile.
     */
    function GetDistanceManhattanToTile(tile)
    {
        return this._API.GetDistanceManhattanToTile(this.GetID(), tile);
    }

    /**
     *  Get an area around servable object
     * @return array of tiles around
     */
    function GetArea() { return this._storage._area ; }

    /**
     *  Set an area around servable object
     * @param array of tiles around
     */
    function SetArea(anArea) { this._storage._area = anArea; }

    /**
     *  Just make sure it was valid
     * @return true if it was valid
     */
    function IsValid()
    {
        return this.IsTown() ? AITown.IsValidTown(this.GetID()) : AIIndustry.IsValidIndustry(this.GetID());
    }

    /**
     * New Servable Town
     */
    static function NewTown(id)
    {
        local tmp = Servable();
        tmp.SetID(id);
        tmp._API = AITown;
        local area = Tiles.OfTown(tmp.GetID(), (Tiles.Radius(tmp.GetLocation(), 20)));
        tmp.SetArea(Assist.ListToArray(area));
        return tmp;
    }

    /**
     * New Servable Industry
     */
    static function NewIndustry(id)
    {
        local tmp = Servable();
        tmp.SetID(id);
        tmp._API = AIIndustry;
        tmp._storage._is_town <- false; ///< Flag of town
        local area =  AITileList_IndustryProducing(tmp.GetID(), 10);
        if (area.IsEmpty()) area = AITileList_IndustryAccepting(tmp.GetID(), 10);
        tmp.SetArea(Assist.ListToArray(area));
        return tmp;
    }

    /**
     * New Unknown Servable
     */
    static function New(id, istown)
    {
        if (istown) return Servable.NewTown(id);
        return Servable.NewIndustry(id);
    }

    /**
     * Make class from storage
     * @param storage to build the class from
     */
    static function MakeClass(storage)
    {
        assert(typeof(storage) == "table");
        local tmp = Servable();
        if (!storage._is_town) tmp._API = AIIndustry;
        tmp.SetStorage(storage);
        return tmp;
    }
}

class ServableManager extends Manager
{
	constructor()
	{
		::Manager.constructor("Servable");
	}
	
	/**
     * Overriden to Get a class from list
     * @param item ID
     * @return clss of item
     */
	function Item(id)
	{
		return Servable.MakeClass(this._storage.rawget(id)); 
	}
	
	/**
	 * Insert New Servable Item
	 * @param serv Servable class to insert
	 * @return Available ID
	 */
	 function New(serv)
	 {
	 	assert(serv instanceof Servable);
        local c = 1;
        while (this.list.HasItem(c)) c++;
        local val = serv.GetID() + serv.GetLocation();
        this.Push(c, val, serv.GetStorage());
        return c;
	 }
	 
	 /**
	 * Find an ID using key of servable
	 * @param key Key to find
	 * @param istown Wether to find town or industry
	 * @return a valid ID or null if not found
	 */
	function FindKey(key, istown)
	{
		local v1 = AITown.GetLocation(key);
		local v2 = AIIndustry.GetLocation(key);
		v1 = AIMap.IsValidTile(v1) ? (v1 + key) : 0;
		v2 = AIMap.IsValidTile(v2) ? (v2 + key) : 0;
		if (v1 || v2) {
			foreach (idx, val in this.list)	{
				if ((val == v1) || (val == v2)) {
					if (this._storage.rawget(idx)._id == key) return idx;
				}
			}
		}
	}
}
