/*  09.05.22 - task.nut
 *
 *  This file is part of Trans AI
 *
 *  Copyright 2009 fanio zilla <fanio.zilla@gmail.com>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 *  MA 02110-1301, USA.
 */

/**
 * Fake Task Namespace
 */
class Task {}

/**
 * Task item is base for all task item
 */
class TaskItem
{
	_removable = null;
	_taskname = null;
	_result = null;

	constructor(name)
	{
		this._removable = true;
		this._taskname = name;
		this._result = null;
	}

	/**
	 * Set task removability
	 * @param val Fill true if removable
	 * @note Maybe we don't need this
	*/
	function SetRemovable(val) { this._removable = val; }

	/**
	 * Is this task removable
	 * @return true if we should delete this task after execute it
	*/
	function IsRemovable() { return this._removable }

	/**
	 * Execute this task
	 * @note to be overriden by class descendants
	 * @return null if not execute anything
	 */
	function Execute()
	{
		AILog.Warning("Not defined Execute method yet");
	}

	/**
	 * Get the result of executing this task
	 * @return anything
	 */
	function GetResult() { return this._result; }

	/**
	 * Get the name of task
	 * @return task name
	 */
	function GetName() {return this._taskname; }
}

/**
 * Task that is using yield from generator
*/
class YieldTask extends TaskItem
{
	///last yield holder
	last_yield = null;
	/// flag to repeat generator
	_repeat = null;
	
	constructor(name)
	{
		::TaskItem.constructor(name);
		this.last_yield = "new";
		this._repeat = true;
		this._removable = false;
	}

	function Execute()
	{
		if (this.last_yield == "new") this.last_yield = this._exec();
		try {
			this._result = resume this.last_yield;
			return this._result;
		}
		catch (last_yield) {
			AILog.Info("Generator stopped");
		}
		if (this._repeat) {
			AILog.Info("Rebuild yield");
			this.last_yield = "new";
			return true;
		}
		this._result = "stop";
		
	}

	function _exec()
	{
		AILog.Warning("Not defined _exec method yet");
	}
}

/**
 * Task management
 */
class TaskManager extends KeyLists
{
	/** local  ticker */
	tick =  null;
	/** local  ticker */
	_sleep_time = null;

	constructor()
	{
		::KeyLists.constructor();
		tick = 0;
		this._sleep_time = 10;
	}

	/**
	 * Insert new task into scheduler
	 * @param task Class of TaskItem to insert
	 * @param looper In what tick will executed
	 * @return the ID of this task in scheduler
	 */
	function New(task, looper)
	{
		assert(task instanceof TaskItem);
		local c = 1;
		while (this.list.HasItem(c)) c++;
		this.Push(c, looper, task);
		return c;
	}

	/**
	 * Set Sleep time for each iteration
	 * @param time Time in tick to sleep
	 */
	function SetSleep(time)
	{
		this._sleep_time = time;
	}

	/**
	 * Run the scheduler
	 * execute scheduled task by tick
	 */
	function Run () {
		if (Debug.ResultOf("Our Task" , this.Count()) == 0) return;
		this.tick ++;
		local task = null;
		//local msg = "[" + this.tick + "]";
		this.SortValueAscending();
		for (local i= this.list.Begin(); this.list.HasNext(); i = this.list.Next()) {
			AIController.Sleep(this._sleep_time);
			task = this.Item(i);
			if (this.tick % this.GetValue(i) == 0) {
				Debug.ResultOf(/* msg + "Run:" + */ task.GetName(), task.Execute());
				if (task instanceof YieldTask) {
					if (task.GetResult() != "stop") continue;
				}

				if (task.IsRemovable()) this.RemoveItem(i);
			} else {
				//AILog.Info(msg + "Suspend:" + task.GetName());
			}
			task = null;
		}
		if (this.tick == 0xFFFF) this.tick = 0;
	}

	/**
	 * Find a task by matching its name *
	 * @param name Name of task
	 * @return the ID of this task in scheduler or null if not found
	 */
	function FindByName(name)
	{
		foreach (idx, val in this.list) if (this.Item(idx).GetName() == name) return idx;
	}
}
