/*  09.02.04 - sandbox.nut
 *
 *  This file is part of Trans AI
 *
 *  Copyright 2009 fanio zilla <fanio.zilla@gmail.com>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 *  MA 02110-1301, USA.
 */

 /**
 * General un categorized static functions to assist program
 */
class Assist
{
	/**
	 * Sleep Time
	 * @return The amount time for company to sleep :-)
	 */
	function SleepTime()
	{
		if (AICompany.GetLoanAmount() < 20000) return 20;
		return (AICompany.GetLoanAmount() / 500).tointeger();
	}

	/**
	 * Rename the drop off station for industry
	 *@param st_id Station ID
	 *@param ind_id Industry ID
	 */
	static function RenameStation(st_id, ind_id)
	{
		local counter = 1;
		local name = AIIndustryType.GetName(AIIndustry.GetIndustryType(ind_id)) + " " + this._commander.randomizer;
		while (!AIStation.SetName(st_id, name + " Drop Off " + counter) && counter < 100) counter++;
	}

	/**
	 * ClearSigns is AISign cleaner
	 * Clear all sign that I have been built while servicing.
	 */
	static function ClearSigns()
	{
		AILog.Info("Clearing signs ...");
		for (local c = AISign.GetMaxSignID(); c > 0; c--) if (AISign.IsValidSign(c)) AISign.RemoveSign(c);
	}

	/**
	 * Count cargo that accept by an Industry
	 * @param id Industry ID
	 * @return number of cargo
	 */
	static function CargoCount(id)
	{
		local ret = 0;
		local type = AIIndustry.GetIndustryType(id);
		local cargoes = AIIndustryType.GetAcceptedCargo(type);
		if (cargoes) {
			ret += cargoes.Count();
		}
		return ret;
	}

	/**
	 * Temporary service cost
	 * @param src Source ID (industry/town)
	 * @param istown Fill true if this is town
	 * @param cargo ID of cargo to be used
	 */
	static function ServiceCost(src, istown, cargo)
	{
		local cost = 0;
		local multiplier = 1;
		local fnAPI = istown ? AITown : AIIndustry ;
		if (istown) cost += AITown.GetMaxProduction(src, cargo)
		else cost += fnAPI.GetLastMonthProduction(src, cargo);
		cost -= fnAPI.GetLastMonthTransported(src, cargo) * multiplier;
		cost +=  AICargo.GetCargoIncome(cargo, 20, 200);
		return cost;
	}

	/**
	 * Average an integer list in array
	 * @param parray Array of integer number
	 * @return Average of number
	 */
	static function Average(parray)
	{
		assert(parray);
		local sam = 0, count = parray.Count();
		foreach (idx, val in parray) sam += val;
		local mod = sam % count;
        local rounded = (mod / count) > 0.5 ? 1: 0;
		return  (sam - mod) / count + rounded;
	}

	/**
	 * Valuator function
	 * @param list AIList to valuate
	 * @param valuator Function to be used as valuator
	 * @param ... Additional argument to be passed to
	 */
	 static function Valuate(list, valuator, ...)
	{
		assert(list);
		assert(list instanceof AIAbstractList);
		assert(typeof(valuator) == "function");

		local args = [this, null];
		for(local c = 0; c < vargc; c++) args.append(vargv[c]);
		foreach(idx, val in list) {
			args[1] = idx;
			local value = valuator.acall(args);
			if (typeof(value) == "bool") {
				value = value ? 1 : 0;
			} else if (typeof(value) != "integer") throw("Invalid return type from valuator");
			list.SetValue(idx, value);
		}
	}

	/**
	 * Check if parameter is null
	 * @param val val to evaluate
	 * @return true if val is null
	 */
	static function IsNull(val) { return val == null ; }

	/**
     * Lead a number with zero
     * this will solve problem of '09' that displayed '9' only
     * @param integer_number to convert
     * @return number in string
     * @note only for number below 10
     */
    static function LeadZero(integer_number)
    {
        if (integer_number > 9) return integer_number.tostring();
        return "0" + integer_number;
    }

    /**
     * Convert date to it string representation in DD-MM-YYYY
     * @param date to convert
     * @return string representation in DD-MM-YYYY
     */
    static function DateStr(date)
    {
        return "" + AIDate.GetDayOfMonth(date) + "-" +   AIDate.GetMonth(date) + "-" + AIDate.GetYear(date);
    }

    /**
     * Hex to Decimal converter
     * @param Hex_number in string to convert
     * @return  number in integer
     * @note max number is 255 or FF
    */
    static function HexToDec(Hex_number)
    {
        if (Hex_number.length() > 2) return 0;
        local aSet = "0123456789ABCDEF";
        return aSet.find(Hex_number.slice(0,1)).tointeger() * 16 + aSet.find(Hex_number.slice(1,2)).tointeger();
    }

	/**
     * Decimal to Hex converter
     * @param dec number to convert
     * @return hex number in string
    */
    static function DecToHex(dec)
	{
		local aSet = ["0","1","2","3","4","5","6","7","8","9","A","B","C","D","E","F"];
		local tmp = [];
		local c = dec % 16;
		while (true) {
			tmp.push(aSet[c]);
			if (dec < 16) break;
			dec = (dec - c) / 16;
			c = dec % 16;
		}
		if (tmp.len() == 1) tmp.push("0");
		tmp.reverse();
		local ret = "";
		foreach (idx, val in tmp) ret += val;
		return ret;
	}

    /**
     * Push list to array
     * @param list AIList to convert from
     * @return Array of index list
     */
    static function ListToArray(list)
    {
        local array = [];
        foreach(item, lst in list) array.push(item);
        return array;
    }

    /**
     * Make list from array
     * @param array to convert from
     * @return AIList of array value
     */
    static function ArrayToList(array)
    {
        local list = AIList();
        foreach(idx, item in array) list.AddItem(item, 0);
        return list;
    }

	/**
	 * Convert path to array
	 * @param path Path class to convert from
     * @return Array of path.GetTile()
     */
	static function Path2Array(path)
	{
		local anArray = [];
		local cur = -1;
		while (path != null) {
			anArray.push(path.GetTile());
			path = path.GetParent();
		}
		return anArray; //.reverse();
	}

    /**
	 * Filter Industry that is built on water
	 * @param anIndustry ID of industry
     * @return AIList of non water Industry
     */
     static function NotOnWater(anIndustry) {
		anIndustry.Valuate(AIIndustry.IsValidIndustry);
		anIndustry.RemoveValue(0);
        anIndustry.Valuate(AIIndustry.IsBuiltOnWater);
        anIndustry.RemoveValue(1);
        return anIndustry;
    }

    /**
     * Additional cost for road path finder
     * @param path Path class
     * @param new_tile The next tile
     * @param new_direction The next direction
     * @return integer cost
     */
    static function RoadDiscount(path, new_tile, new_direction)
    {
        local new_cost = 0;
        local prev_tile = (path.GetParent() == null) ? new_tile : path.GetParent().GetTile();
        if (AIBridge.IsBridgeTile(new_tile) && (AIBridge.GetOtherBridgeEnd(new_tile) == prev_tile)) {
            local b_id = AIBridge.GetBridgeID(new_tile);
            if (AIBridge.IsValidBridge(b_id)) new_cost -= (AIBridge.GetMaxSpeed(b_id)  + this._cost_bridge_per_tile);
        }
        if (AITunnel.IsTunnelTile(new_tile) && AITunnel.GetOtherTunnelEnd(new_tile) == prev_tile) {
            new_cost -= this._cost_tunnel_per_tile;
        }
        if (AIRoad.IsRoadTile(new_tile) && AIRoad.AreRoadTilesConnected(prev_tile, new_tile)) {
            new_cost -= this._cost_tile * 2;
        }

        if (!AITile.DemolishTile(new_tile) && AIError.GetLastError() == AIError.ERR_VEHICLE_IN_THE_WAY) {
            new_cost += this._cost_bridge_per_tile * 2;
        }
        //AILog.Info("cost:" + new_cost);
        return new_cost * AIMap.DistanceManhattan(new_tile, prev_tile);
    }

    /**
     * @deprecated Deprecated
     */
    static function Connect_BackBone(main, serv_id)
    {
        local _cost = 0;
        local serv = null;
        if (main._commander.service_tables.rawin(serv_id)) {
            serv = main._commander.service_tables[serv_id];
        }
        if (serv.VehicleType != AIVehicle.VT_RAIL) return true;
        AILog.Info("Try to connect backbone for id " + serv_id);
        main.Rail.Path(serv, 2, true);
        main.State.TestMode = true;

        if (!main.Rail.Track(serv, 2)) {
            main.Rail.Path(serv, 2, true);
            if (!main.Rail.Track(serv, 2)) return false;
        }
        _cost += main.State.LastCost;
        main.Rail.Vehicle(serv);
        _cost += main.State.LastCost;
        main.Rail.Signal(serv, 1);
        _cost += main.State.LastCost;
        main.Rail.Signal(serv, 2);
        _cost += main.State.LastCost;
        if (!Bank.Get(_cost)) return false;

        main.State.TestMode = false;
        if (!main.Rail.Track(serv, 2)) {
            main.Rail.Path(serv, 2, true);
            if (!main.Rail.Track(serv, 2)) return false;
        }
        main.Rail.Signal(serv, 1);
        main.Rail.Signal(serv, 2);
        main.Rail.Vehicle(serv);
        return true;
    }

	/**
	 * Get maximum number of tile acceptance/production
	 * @return integer_number
	 */
    static function GetMaxProd_Accept(tiles, cargo, is_source)
    {
        local check_fn = is_source ? AITile.GetCargoProduction : AITile.GetCargoAcceptance;
		tiles.Valuate(check_fn, cargo, 1, 1, Stations.RoadRadius());
        return check_fn(tiles.Begin(), cargo, 1, 1, Stations.RoadRadius());
    }
}


/**
 * Debug static functions class
 *
 */
class Debug
{
    /**
     * Evaluate expression, display message,  detect last error.
     * usable for in-line debugging
     * @param msg Message to be displayed
     * @param exp Expression to be displayed and returned
     * @return Value of expression
     */
    static function ResultOf(msg, exp)
    {
        if (AIError.GetLastError() == AIError.ERR_NONE) AILog.Info("" + msg + ":" + exp +" -> Good Job :-D");
        else AILog.Warning("" + msg + ":" + exp + ":" + AIError.GetLastErrorString().slice(4));
        /* no other methode found to clear last err */
        //AISign.RemoveSign(Debug.Sign(AIMap.GetTileIndex(2, 2), "debugger"));
        return exp;
    }

    /**
     * The function that should never called / passed by flow of code.
     * @param msg The message to be displayed
	 * @param suspected The variable to displayed
	 * @note This would only set to make the AI end it's live
     */
    static function DontCallMe(msg, suspected)
    {
        /* I've said, don't call me. So why you call me ?
         * okay, I'll throw you out ! :-( */
        AILog.Warning("Should not come here!" + msg + (suspected ? " suspected --> " + suspected : ""));
        //TransAI.Memory.Live = 0;
		throw msg;
    }

    /**
	 * Wrapper for build sign.
	 * prepared for use with Game.Settings
	*/
	static function Sign(tile, txt)
    {
        if (1 == 1) return AISign.BuildSign(tile, txt);
    }

	/**
	 * Assert replacement for custom handling
	 */
	static function Assert(exp)
	{
		if (exp) return true;
		Debug.DontCallMe("Try call null:", exp);
	}
}

class Settings
{
    /* usage : AILog.Info(Const.Settings.long_train + " -> " +  Settings.Get(Const.Settings.long_train)) */
    static function Get(setting_str)
    {
        local temp = AIGameSettings.GetValue(setting_str);
        return (temp == -1) ? false : temp;
    }

	static function Version()
	{
		local v = AIController.GetVersion();
		local maj = (v && 0xF0000000) >> 28;
		local min = (v && 0x0F000000) >> 24;
		local bld = (v && 0x00F00000) >> 20;
		local rel = (v && 0x00080000) != 0;
		local rev = v && 0x0007FFFF;
		if (maj < 1) maj = 0;
		AILog.Info("Running Ver:" + maj + "." + min + " Build:" + bld + " (" + (rel ?  "Release" : "Rev." + rev) + ")");
	}
}

class Ticker
{
	_tick = null; ///< the ticker
	constructor(){
		this._tick = AIController.GetTick();
	}

	function Elapsed() {return AIController.GetTick() - this._tick; }
	function Reset() { this._tick = AIController.GetTick(); }
}
